// Package httpcli is http request client, which only supports return json format.
package httpcli

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"gitlab.wanzhuangkj.com/tush/xpkg/httpcli/entity"
	"io"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/duke-git/lancet/v2/retry"
	"gitlab.wanzhuangkj.com/tush/xpkg/gin/ctxUtils"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
)

const defaultTimeout = 30 * time.Second

type Request struct {
	url         string
	method      string
	headers     map[string]string
	params      map[string]interface{} // parameters after URL
	reqBody     []byte                 // Body data
	reqBodyJSON interface{}            // 传入对象，会自动json marshal
	timeout     time.Duration          // Client timeout
	retryCount  uint

	request  *http.Request
	response *Response
	err      error
}

// Response HTTP response
type Response struct {
	body       []byte
	StatusCode int
	Status     string
	Header     http.Header
}

// -----------------------------------  Request way 1 -----------------------------------

// New create a new Request
func New() *Request {
	return &Request{
		response: &Response{},
	}
}

// Reset set all fields to default value, use at pool
func (x *Request) Reset() {
	x.params = nil
	x.reqBody = nil
	x.reqBodyJSON = nil
	x.timeout = 0
	x.headers = nil

	x.response = nil
	x.method = ""
	x.err = nil
}

// SetURL set URL
func (x *Request) SetURL(path string) *Request {
	x.url = path
	return x
}

// SetParams parameters after setting the URL
func (x *Request) SetParams(params map[string]interface{}) *Request {
	if x.params == nil {
		x.params = params
	} else {
		for k, v := range params {
			x.params[k] = v
		}
	}
	return x
}

// SetParam parameters after setting the URL
func (x *Request) SetParam(k string, v interface{}) *Request {
	if x.params == nil {
		x.params = make(map[string]interface{})
	}
	x.params[k] = v
	return x
}

// SetBody set body data, support string and []byte, if it is not string, it will be json marshal.
func (x *Request) SetBody(body interface{}) *Request {
	switch v := body.(type) {
	case string:
		x.reqBody = []byte(v)
	case []byte:
		x.reqBody = v
	default:
		x.reqBodyJSON = body
	}
	return x
}

// SetTimeout set timeout
func (x *Request) SetTimeout(t time.Duration) *Request {
	x.timeout = t
	return x
}
func (x *Request) SetRetry(count uint) *Request {
	x.retryCount = count
	return x
}

// SetContentType set ContentType
func (x *Request) SetContentType(a string) *Request {
	x.SetHeader("Content-Type", a)
	return x
}

// SetHeader set the value of the request header
func (x *Request) SetHeader(k, v string) *Request {
	if x.headers == nil {
		x.headers = make(map[string]string)
	}
	x.headers[k] = v
	return x
}

// SetHeaders set the value of Request Headers
func (x *Request) SetHeaders(headers map[string]string) *Request {
	if x.headers == nil {
		x.headers = make(map[string]string)
	}
	for k, v := range headers {
		x.headers[k] = v
	}
	return x
}

// GET send a GET request
func (x *Request) GET(ctx context.Context) *Request {
	x.method = http.MethodGet
	x.pull(ctx)
	return x
}

// DELETE send a DELETE request
func (x *Request) DELETE(ctx context.Context) *Request {
	x.method = http.MethodDelete
	x.pull(ctx)
	return x
}

// POST send a POST request
func (x *Request) POST(ctx context.Context) *Request {
	x.method = http.MethodPost
	x.push(ctx)
	return x
}

// PUT send a PUT request
func (x *Request) PUT(ctx context.Context) *Request {
	x.method = http.MethodPut
	x.push(ctx)
	return x
}

// PATCH send PATCH requests
func (x *Request) PATCH(ctx context.Context) *Request {
	x.method = http.MethodPatch
	x.push(ctx)
	return x
}

// Do a request
func (x *Request) Do(ctx context.Context, method string, data interface{}) *Request {
	x.method = method
	switch method {
	case http.MethodGet, http.MethodDelete:
		if data != nil {
			if params, ok := data.(map[string]interface{}); ok { //nolint
				x.SetParams(params)
			} else {
				x.err = errors.New("params is not a map[string]interface{}")
				return x
			}
		}
		x.pull(ctx)
	case http.MethodPost, http.MethodPut, http.MethodPatch:
		if data != nil {
			x.SetBody(data)
		}
		x.push(ctx)
	default:
		x.err = errors.New("unknow method " + method)
	}
	return x
}

func (x *Request) pull(ctx context.Context) {
	val := ""
	if len(x.params) > 0 {
		values := url.Values{}
		for k, v := range x.params {
			values.Add(k, fmt.Sprintf("%v", v))
		}
		val += values.Encode()
	}

	if val != "" {
		if strings.Contains(x.url, "?") {
			x.url += "&" + val
		} else {
			x.url += "?" + val
		}
	}
	x.send(ctx)
}

func (x *Request) push(ctx context.Context) {
	if x.reqBodyJSON != nil {
		body, err := json.Marshal(x.reqBodyJSON)
		if err != nil {
			x.err = err
			return
		}
		x.reqBody = body
	}
	x.send(ctx)
}

func (x *Request) send(ctx context.Context) {
	bodyBuf := bytes.NewBuffer(x.reqBody)
	x.request, x.err = http.NewRequest(x.method, x.url, bodyBuf)
	if x.err != nil {
		return
	}
	if x.reqBody != nil {
		if c, _ := ctxUtils.GetGinCtx(ctx); c != nil {
			entity.SetCopyApiReq(c, bodyBuf)
		}
	}
	if x.headers != nil {
		for k, v := range x.headers {
			x.request.Header.Add(k, v)
		}
	}
	x.request.Header.Add(ctxUtils.HeaderXRequestIDKey, ctxUtils.CtxRequestID(ctx))
	if x.timeout < 1 {
		x.timeout = defaultTimeout
	}
	ctx, cancel := context.WithTimeout(ctx, x.timeout)
	defer cancel()
	x.request.WithContext(ctx)

	logger.Info("[httpCli] req",
		logger.String("method", x.method),
		logger.String("url", x.url),
		logger.Any("header", x.request.Header),
		logger.String("body", bodyBuf.String()),
		ctxUtils.CtxTraceIDField(ctx))

	st := time.Now()
	if x.retryCount > 0 {
		x.err = retry.Retry(func() error {
			x.pushDo(ctx)
			return x.err
		},
			retry.RetryTimes(x.retryCount),
			retry.RetryWithLinearBackoff(8*time.Second),
			retry.Context(ctx),
		)
	} else {
		x.pushDo(ctx)
	}

	logger.Info("[httpCli] rsp",
		logger.String("cost", time.Since(st).String()),
		logger.Int("statusCode", x.response.StatusCode),
		logger.String("status", x.response.Status),
		logger.Any("header", x.Response().Header),
		logger.String("body", x.RespBodyString()),
		ctxUtils.CtxTraceIDField(ctx))
}

func (x *Request) pushDo(ctx context.Context) {
	client := http.Client{Timeout: x.timeout}
	var rsp *http.Response
	rsp, x.err = client.Do(x.request)
	if x.err != nil {
		x.request = nil
		return
	}
	if rsp != nil {
		x.response.Status = rsp.Status
		x.response.Header = rsp.Header
		x.response.StatusCode = rsp.StatusCode
		ctxUtils.SetApiCost(ctx, x.response.Header)
		if rsp.Body != nil {
			defer rsp.Body.Close()
			x.response.body, x.err = io.ReadAll(rsp.Body)
			if x.err != nil {
				x.request = nil
				return
			}
			if len(x.response.body) > 0 {
				if c, _ := ctxUtils.GetGinCtx(ctx); c != nil {
					entity.SetCopyApiRsp(c, x.response.BodyBuf())
				}
			}
		}
	}
	x.request = nil
}

// -----------------------------------  Response -----------------------------------

func (x *Request) Err() error {
	return x.err
}

// RespBodyString returns the body data of the HttpResponse
func (x *Request) RespBodyString() string {
	if len(x.response.body) == 0 {
		return ""
	}
	return bytes.NewBuffer(x.response.body).String()
}
func (x *Request) Response() *Response {
	return x.response
}
func (x *Response) BodyBuf() *bytes.Buffer {
	return bytes.NewBuffer(x.body)
}

func (x *Request) ReadRspBody() ([]byte, error) {
	if x.err != nil {
		return []byte{}, x.err
	}

	if x.response.body == nil {
		return []byte{}, errors.New("nil")
	}

	return x.response.body, nil
}

// BindJSON parses the response's body as JSON
func (x *Request) BindJSON(v interface{}) *Request {
	if x.err != nil {
		return x
	}
	if err := json.Unmarshal(x.response.body, v); err != nil {
		x.err = err
	}
	return x
}

// -----------------------------------  Request way 2 -----------------------------------

// Option set options.
type Option func(*options)

type options struct {
	params  map[string]interface{}
	headers map[string]string
	timeout time.Duration
}

func (o *options) apply(opts ...Option) {
	for _, opt := range opts {
		opt(o)
	}
}

func defaultOptions() *options {
	return &options{}
}

// WithParams set params
func WithParams(params map[string]interface{}) Option {
	return func(o *options) {
		o.params = params
	}
}

// WithHeaders set headers
func WithHeaders(headers map[string]string) Option {
	return func(o *options) {
		o.headers = headers
	}
}

// WithTimeout set timeout
func WithTimeout(t time.Duration) Option {
	return func(o *options) {
		o.timeout = t
	}
}

//
//// Get request, return custom json format
//func Get(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return gDo(ctx, "GET", result, urlStr, o.params, o.headers, o.timeout)
//}
//
//// Delete request, return custom json format
//func Delete(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return gDo(ctx, "DELETE", result, urlStr, o.params, o.headers, o.timeout)
//}
//
//// Post request, return custom json format
//func Post(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return do(ctx, "POST", result, urlStr, body, o.params, o.headers, o.timeout)
//}
//
//// Put request, return custom json format
//func Put(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return do(ctx, "PUT", result, urlStr, body, o.params, o.headers, o.timeout)
//}
//
//// Patch request, return custom json format
//func Patch(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return do(ctx, "PATCH", result, urlStr, body, o.params, o.headers, o.timeout)
//}
//
//var requestErr = func(err error) error { return fmt.Errorf("request error, err=%v", err) }
//var jsonParseErr = func(err error) error { return fmt.Errorf("json parsing error, err=%v", err) }
//var notOKErr = func(resp *Response) error {
//	body, err := resp.ReadBody()
//	if err != nil {
//		return err
//	}
//	if len(body) > 500 {
//		body = append(body[:500], []byte(" ......")...)
//	}
//	return fmt.Errorf("statusCode=%d, body=%s", resp.StatusCode, body)
//}
//
//func do(ctx context.Context, method string, result interface{}, urlStr string, body interface{}, params KV, headers map[string]string, timeout time.Duration) error {
//	if result == nil {
//		return fmt.Errorf("'result' can not be nil")
//	}
//
//	req := &Request{}
//	req.SetURL(urlStr)
//	req.SetContentType("application/json")
//	req.SetParams(params)
//	req.SetHeaders(headers)
//	req.SetBody(body)
//	req.SetTimeout(timeout)
//
//	var resp *Response
//	var err error
//	switch method {
//	case "POST":
//		resp, err = req.POST(ctx)
//	case "PUT":
//		resp, err = req.PUT(ctx)
//	case "PATCH":
//		resp, err = req.PATCH(ctx)
//	}
//	if err != nil {
//		return requestErr(err)
//	}
//	defer resp.Body.Close() //nolint
//
//	if resp.StatusCode != 200 {
//		return notOKErr(resp)
//	}
//
//	err = resp.BindJSON(result)
//	if err != nil {
//		return jsonParseErr(err)
//	}
//
//	return nil
//}

//
//func gDo(ctx context.Context, method string, result interface{}, urlStr string, params KV, headers map[string]string, timeout time.Duration) error {
//	req := &Request{}
//	req.SetURL(urlStr)
//	req.SetParams(params)
//	req.SetHeaders(headers)
//	req.SetTimeout(timeout)
//
//	var resp *Response
//	var err error
//	switch method {
//	case "GET":
//		resp, err = req.GET(ctx)
//	case "DELETE":
//		resp, err = req.DELETE(ctx)
//	}
//	if err != nil {
//		return requestErr(err)
//	}
//	defer resp.Body.Close() //nolint
//
//	if resp.StatusCode != 200 {
//		return notOKErr(resp)
//	}
//
//	err = resp.BindJSON(result)
//	if err != nil {
//		return jsonParseErr(err)
//	}
//
//	return nil
//}

// StdResult standard return data
type StdResult struct {
	Code int         `json:"code"`
	Msg  string      `json:"msg"`
	Data interface{} `json:"data,omitempty"`
}

// StdResult standard return data
type StdResult2 struct {
	Code int         `json:"code"`
	Msg  string      `json:"message"`
	Data interface{} `json:"data,omitempty"`
}

// KV string:interface{}
type KV = map[string]interface{}
